import Control.Applicative
import Control.Monad
import Control.Parallel.Strategies
import qualified Control.Exception as E
import Impl
import HexpatImpl
import HaXmlImpl
import HXTImpl
import LightImpl
import Data.IORef
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as L
import Data.ByteString.Internal (c2w, w2c)
import System.Directory
import System.Posix.Files
import Data.Maybe
import Data.Ord
import Data.List
import System.Environment
import System.Random
import System.CPUTime
import System.IO
import Debug.Trace


data Input = Input {
        inputText :: IORef InputText,
        inputName :: String,
        inputSize :: Int
    }

main = do
    let outerLoops = 5
    args <- getArgs
    let (title, measureForUs) = case args of
            [title, mf] -> (title, read mf)
            otherwise -> error "Usage: benchmark <title> <no of us to measure for>"
    putStrLn title
    files <- getDirectoryContents "files/"
    inputs <- sortBy (comparing inputSize) . catMaybes <$> forM files (\file0 -> do
        let file = "files/"++file0
        fs <- getFileStatus file
        if isRegularFile fs
            then do
                xml <- L.readFile file
                let it = InputText {
                            textBS  = B.concat $ L.toChunks xml,
                            textLBS = xml,
                            textS   = map w2c $ L.unpack xml
                        }
                rnf it `seq` return ()
                itRef <- newIORef it 
                return $ Just $ Input {
                        inputText = itRef,
                        inputName = file,
                        inputSize = fromIntegral $ fileSize fs
                    }
            else return Nothing)
    let parsers = [haXmlImpl] ++ [lightImpl] ++ {-[hxtImpl] ++-} hexpatImpls
        tTotal = fromIntegral (length parsers) * fromIntegral (length inputs) * fromIntegral outerLoops * measureForUs :: Integer
    hPutStrLn stderr $ "Will take approx " ++ show (realToFrac tTotal / 60000000.0) ++ " minutes"
    resultsTables <- forM [1..outerLoops] $ \loop -> do
        table <- forM inputs $ \input -> do
            row <- forM parsers $ \parser ->
                do
                    (duration, iters) <- bench measureForUs $ do
                        txt <- readIORef (inputText input)
                        parseFunc parser txt
                    return $ duration `div` fromIntegral iters
                 `E.catch` (\exc -> do
                      putStrLn $ (inputName input) ++ " " ++ show (exc :: E.SomeException)
                      return 0
                    )
            hPutStrLn stderr $ inputName input ++ " "++show row
            return row
        hPutStrLn stderr ""
        hPutStrLn stderr $ title++" " ++ show (loop * 20) ++ "%"
        hPutStrLn stderr ""
        return table
    let results = overlay resultsTables
    print $ map libname parsers
    forM_ (zip results inputs) $ \(row, input) -> do
        print $ (inputSize input, map (median . map (`div` 1000000)) row)  -- convert to us

bench :: Integer -> IO () -> IO (Integer, Int)
bench maxUs action = do
    start <- getCPUTime
    let desiredEnd = start + maxPico
    iters <- bench' desiredEnd 0
    end <- getCPUTime
    return (end - start, iters)
  where
    maxPico = maxUs * 1000000
    bench' desiredEnd iters = do
        forM [1..16] $ \_ -> action
        t <- getCPUTime
        let iters' = iters + 16
        if t < desiredEnd
            then bench' desiredEnd iters'
            else return iters'

-- | Combine a set of result tables into a result table where each element is the list of results
-- from each of the result tables.
overlay :: [[[Integer]]] -> [[[Integer]]]
overlay = foldl' o1 (repeat (repeat []))
  where
    o1 :: [[[Integer]]] -> [[Integer]] -> [[[Integer]]]
    o1 = zipWith $ zipWith $ flip (:)

median :: [Integer] -> Integer
median ns = sort ns !! (length ns `div` 2)

